@@ -106,10 +106,12 @@ def _configure_autoscaler_credentials(
 
 @_with_safe_cluster_config  # type: ignore
 def _update_file_mounts(
-    cluster_config: Dict[str, Any], project_root_dir: str
+    cluster_config: Dict[str, Any],
+    project_root_dir: str,
+    api_client: Optional[DefaultApi] = None,
 ) -> Dict[str, Any]:
     project_id = get_project_id(project_root_dir)
-    working_dir = get_working_dir(cluster_config, project_id)
+    working_dir = get_working_dir(cluster_config, project_id, api_client)
     for remote_path, local_path in cluster_config.get("file_mounts", {}).items():
         if remote_path == working_dir and not os.path.samefile(
             local_path, project_root_dir
@@ -188,7 +190,9 @@ def configure_for_session(
         cluster_config, session_id, api_client
     )
     if not disable_project_sync:
-        cluster_config = _update_file_mounts(cluster_config, project_root_dir)
+        cluster_config = _update_file_mounts(
+            cluster_config, project_root_dir, api_client
+        )
 
     # This is to prevent autoscaler from creating a new security group.
     cluster_config["provider"].update(
