@@ -229,7 +229,7 @@ class CloudController(object):
         print(self.get_cloud_config(cloud_name=None, cloud_id=cloud_id))
 
     def get_cloud_config(
-        self, cloud_name: Optional[str], cloud_id: Optional[str],
+        self, cloud_name: Optional[str] = None, cloud_id: Optional[str] = None,
     ) -> str:
         """Get a cloud's current JSON configuration."""
 
