@@ -1,3 +1,4 @@
+import json
 import os
 from typing import Optional
 
@@ -21,6 +22,8 @@ from anyscale.util import (
     validate_cluster_configuration,
 )
 
+COMPUTE_CONFIG_FILENAME = "example_compute_config.json"
+
 
 class ProjectController:
     def __init__(self, api_client: Optional[DefaultApi] = None):
@@ -34,6 +37,10 @@ class ProjectController:
         os.makedirs(project_name)
         clone_cluster_config(project_name, project_name, project_id, self.api_client)
 
+        self._write_sample_compute_config(
+            os.path.join(project_name, COMPUTE_CONFIG_FILENAME)
+        )
+
     def init(
         self,
         name: Optional[str] = None,
@@ -110,3 +117,19 @@ class ProjectController:
             )
 
         register_project(project_definition, self.api_client)
+        self._write_sample_compute_config(COMPUTE_CONFIG_FILENAME)
+
+    def _write_sample_compute_config(self, filepath: str) -> None:
+        """Writes a sample compute config JSON file to be used with anyscale up.
+        """
+
+        # Compute configs need a real cloud ID;
+        # since this is just a sample file, we can pick any of the user's clouds.
+        cloud_id = self.api_client.list_clouds_api_v2_clouds_get().results[0].id
+
+        default_config = self.api_client.get_default_compute_config_api_v2_compute_templates_default_cloud_id_get(
+            cloud_id=cloud_id
+        ).result
+
+        with open(filepath, "w") as f:
+            json.dump(default_config.to_dict(), f, indent=2)
