@@ -34,6 +34,7 @@ from anyscale.client.openapi_client.api.default_api import DefaultApi  # type: i
 from anyscale.cloud import get_cloud_id_and_name
 from anyscale.cluster_config import (
     _configure_min_max_workers,
+    _update_file_mounts,
     configure_for_session,
     get_cluster_config,
 )
@@ -415,6 +416,15 @@ class SessionController:
             if no_rapid_start:
                 cluster_config["provider"]["bypass_rapid_start"] = True
 
+            if disable_sync:
+                start_empty_head_node_only = False
+                cluster_config["file_mounts"] = {}
+            else:
+                start_empty_head_node_only = True
+                cluster_config = _update_file_mounts(
+                    cluster_config, project_definition.root, self.api_client
+                )
+
         # Build and compute template launch path.
         else:
             # Most of the logic is handled in the backend;
@@ -467,6 +477,7 @@ class SessionController:
                         idle_timeout=idle_timeout,
                         restart_only=restart_only,
                         no_restart=no_restart,
+                        start_empty_head_node_only=start_empty_head_node_only,
                     )
                 ).result
             session_id = up_response.session_id
@@ -476,14 +487,15 @@ class SessionController:
             )
 
             try:
-                # Push will sync project directory to working directory.
-                self.push(
-                    session_name,
-                    source=None,
-                    target=None,
-                    config=None,
-                    all_nodes=False,
-                )
+                if not disable_sync:
+                    # Push will sync project directory to working directory.
+                    self.push(
+                        session_name,
+                        source=None,
+                        target=None,
+                        config=None,
+                        all_nodes=False,
+                    )
             except Exception as e:
                 self.api_client.cancel_startup_api_v2_sessions_session_id_cancel_startup_post(
                     session_id=session_id
