@@ -25,66 +25,9 @@ ANYSCALE_PROJECT_FILE = ".anyscale.yaml"
 ANYSCALE_AUTOSCALER_FILE = "session-default.yaml"
 
 
-CLUSTER_YAML_TEMPLATE = """
-cluster_name: default
-
-max_workers: 1
-
-upscaling_speed: 1.0
-
-docker:
-    image: anyscale/ray-ml:latest
-    container_name: ray_container
-    pull_before_run: true
-
-# Cloud-provider specific configuration.
-provider:
-    type: aws
-    region: us-west-2
-    availability_zone: us-west-2a
-
-available_node_types:
-    anyscale.head.default:
-        node_config:
-            InstanceType: m5.large
-            BlockDeviceMappings:
-              - DeviceName: /dev/sda1
-                Ebs:
-                    VolumeSize: 100
-        resources: {}
-        min_workers: 0
-        max_workers: 0
-    anyscale.worker.small:
-        node_config:
-            InstanceType: m5.large
-        resources: {}
-        min_workers: 0
-        max_workers: 1
-
-# Specify the node type of the head node (as configured above).
-head_node_type: anyscale.head.default
-
-worker_default_node_type: anyscale.worker.small
-
-head_node: {}
-
-worker_nodes: {}
-
-# How Ray will authenticate with newly launched nodes.
-auth:
-    ssh_user: ubuntu
-
-setup_commands: []
-
-head_setup_commands: []
-
-rsync_exclude:
-    - "**/.git"
-    - "**/.git/**"
-
-rsync_filter:
-    - ".gitignore"
-"""
+CLUSTER_YAML_TEMPLATE = open(
+    os.path.join(os.path.dirname(__file__), "default_anyscale_aws.yaml")
+).read()
 
 CLUSTER_CONFIG_TEMPLATE_STR = json.dumps(
     yaml.load(CLUSTER_YAML_TEMPLATE, Loader=yaml.SafeLoader)
@@ -199,7 +142,10 @@ def get_project_sessions(
 
     with format_api_exception(ApiException):
         response = api_client.list_sessions_api_v2_sessions_get(
-            project_id=project_id, name_match=session_name, active_only=True
+            project_id=project_id,
+            name_match=session_name,
+            active_only=True,
+            _request_timeout=30,
         )
     sessions = response.results
     if len(sessions) == 0:
@@ -256,7 +202,7 @@ def _get_project_session(project_id: str, session_name: Optional[str]) -> Any:
 def get_proj_name_from_id(project_id: str, api_client: DefaultApi) -> str:
     with format_api_exception(ApiException):
         resp = api_client.get_project_api_v2_projects_project_id_get(
-            project_id=project_id
+            project_id=project_id, _request_timeout=30
         )
 
     if resp is None:
@@ -275,7 +221,7 @@ def get_proj_id_from_name(
 
     with format_api_exception(ApiException):
         resp = api_client.find_project_by_project_name_api_v2_projects_find_by_name_get(
-            name=project_name
+            name=project_name, _request_timeout=30
         )
 
     if not resp.results:
