@@ -12,11 +12,7 @@ import unicodedata
 from urllib.parse import quote as quote_sanitize, urlencode as urlencode, urljoin
 import webbrowser
 
-import boto3
-import botocore
-from botocore.config import Config
 import click
-import jinja2
 import requests
 from requests import Response
 import yaml
@@ -36,18 +32,22 @@ def confirm(msg: str, yes: bool) -> Any:
     return None if yes else click.confirm(msg, abort=True)
 
 
-def get_endpoint(endpoint: str) -> str:
-    return str(urljoin(anyscale.conf.ANYSCALE_HOST, endpoint))
+def get_endpoint(endpoint: str, host: Optional[str] = None) -> str:
+    return str(urljoin(host or anyscale.conf.ANYSCALE_HOST, endpoint))
 
 
 def send_json_request_raw(
-    endpoint: str, json_args: Dict[str, Any], method: str = "GET"
+    endpoint: str,
+    json_args: Dict[str, Any],
+    method: str = "GET",
+    cli_token: Optional[str] = None,
+    host: Optional[str] = None,
 ) -> Response:
-    if anyscale.conf.CLI_TOKEN is None:
+    if anyscale.conf.CLI_TOKEN is None and not cli_token:
         anyscale.conf.CLI_TOKEN = load_credentials()
 
-    url = get_endpoint(endpoint)
-    cookies = {"cli_token": anyscale.conf.CLI_TOKEN}
+    url = get_endpoint(endpoint, host=host)
+    cookies = {"cli_token": cli_token or anyscale.conf.CLI_TOKEN}
     try:
         if method == "GET":
             resp = requests.get(url, params=json_args, cookies=cookies)
@@ -70,15 +70,21 @@ def send_json_request_raw(
 
 
 def send_json_request(
-    endpoint: str, json_args: Dict[str, Any], method: str = "GET"
+    endpoint: str,
+    json_args: Dict[str, Any],
+    method: str = "GET",
+    cli_token: Optional[str] = None,
+    host: Optional[str] = None,
 ) -> Dict[str, Any]:
-    resp = send_json_request_raw(endpoint, json_args, method=method)
+    resp = send_json_request_raw(
+        endpoint, json_args, method=method, cli_token=cli_token, host=host,
+    )
 
     if not resp.ok:
         if resp.status_code == 500:
             raise click.ClickException(
                 "There was an internal error in this command. "
-                "Please report this to the Anyscale team at https://airtable.com/shrd3BPvtgxR9PxWE "
+                "Please report this to the Anyscale team at beta@anyscale.com "
                 "with the token '{}'.".format(resp.headers["x-trace-id"])
             )
 
@@ -172,18 +178,20 @@ def get_requirements(requirements_path: str) -> str:
 
 
 def _resource(name: str, region: str) -> Any:
+    import boto3
+    from botocore.config import Config
+
     boto_config = Config(retries={"max_attempts": BOTO_MAX_RETRIES})
-    boto_session = boto3.session.Session()
-    return boto_session.resource(name, region, config=boto_config)
+    return boto3.resource(name, region, config=boto_config)
 
 
 def _client(name: str, region: str) -> Any:
-    boto_config = Config(retries={"max_attempts": BOTO_MAX_RETRIES})
-    boto_session = boto3.session.Session()
-    return boto_session.client(name, region, config=boto_config)
+    return _resource(name, region).meta.client
 
 
 def _get_role(role_name: str, region: str) -> Any:
+    import botocore
+
     iam = _resource("iam", region)
     role = iam.Role(role_name)
     try:
@@ -197,6 +205,8 @@ def _get_role(role_name: str, region: str) -> Any:
 
 
 def _get_user(user_name: str, region: str) -> Any:
+    import botocore
+
     iam = _resource("iam", region)
     user = iam.User(user_name)
     try:
@@ -210,8 +220,9 @@ def _get_user(user_name: str, region: str) -> Any:
 
 
 def get_available_regions() -> List[str]:
-    boto_session = boto3.session.Session()
-    client = boto_session.client("ec2")
+    import boto3
+
+    client = boto3.client("ec2")
     return [region["RegionName"] for region in client.describe_regions()["Regions"]]
 
 
@@ -292,6 +303,20 @@ def get_container_name(cluster_config: Dict[str, Any]) -> str:
     return str(cluster_config.get("docker", {}).get("container_name", ""))
 
 
+def get_autoscaler_container_path(path: str) -> str:
+    """
+    When autoscaler is run in a separate container, all locations Docker bind-mounted
+    from the host to directory /a/b/c of the user container are instead bind-mounted
+    to /home/ray/autoscaler_mounts/a/b/c in the autoscaler container.
+
+    Args:
+        path: Location in user container.
+    Returns:
+        Location in the autoscaler container.
+    """
+    return f"/home/ray/autoscaler_mounts/{path}"
+
+
 def get_project_directory_name(project_id: str, api_client: DefaultApi = None) -> str:
     # TODO (yiran): return error early if project doesn't exist.
     with format_api_exception(ApiException):
@@ -320,7 +345,6 @@ def wait_for_session_start(
     if api_client is None:
         api_client = get_api_client()
 
-    retry_count = 0
     while True:
         sessions = api_client.list_sessions_api_v2_sessions_get(
             project_id=project_id, name=session_name, active_only=False
@@ -341,16 +365,20 @@ def wait_for_session_start(
                 raise click.ClickException(
                     f"Error while starting session {session_name}: {session.state_data.startup.startup_error}"
                 )
-            elif session.state and "Errored" in session.state:
+            elif (
+                session.state
+                and "Errored" in session.state
+                and session.pending_state is None
+            ):
                 raise click.ClickException(
                     f"Error while starting session {session_name}: Session startup failed due to an error."
                 )
             elif (
-                retry_count >= 2
-                and session.state
+                session.state
                 and session.state in {"Terminated", "Stopped"}
+                and session.pending_state is None
             ):
-                # Session is created in Terminated state; Ignore the first few state checks.
+                # Session is created in Terminated state; Check pending state to see if it is pending transition.
                 raise click.ClickException(
                     f"Error while starting session {session_name}: Session is still in stopped/terminated state."
                 )
@@ -360,13 +388,19 @@ def wait_for_session_start(
             )
 
         time.sleep(5)
-        retry_count += 1
 
 
 def wait_for_head_node_start(
-    project_id: str, session_name: str, api_client: Optional[DefaultApi] = None
+    project_id: str,
+    session_name: str,
+    session_id: str,
+    api_client: Optional[DefaultApi] = None,
 ) -> str:
-    print(f"Waiting for head node of session {session_name} to start.")
+    url = get_endpoint(f"/projects/{project_id}/sessions/{session_id}")
+    print(
+        f"Waiting for head node of session {session_name} to start. "
+        f"Please view the startup logs at {url}"
+    )
     if api_client is None:
         api_client = get_api_client()
 
@@ -545,6 +579,8 @@ def get_rsync_command(
 
 
 def populate_session_args(cluster_config_str: str, config_file_name: str) -> str:
+    import jinja2
+
     env = jinja2.Environment()
     t = env.parse(cluster_config_str)
     for elem in t.body[0].nodes:
