@@ -11,11 +11,11 @@ from anyscale.client.openapi_client import (  # type: ignore
     Project,
     Session,
     SessionCommand,
+    SessionCommandTypes,
     SessionListResponse,
     SessionStartingUpData,
     SessionStateData,
 )
-from backend.server.base_api.models.session_commands import SessionCommandTypes
 
 
 @pytest.fixture()
