@@ -1,6 +1,6 @@
 import json
 import time
-from unittest.mock import Mock, patch
+from unittest.mock import ANY, Mock, patch
 
 import pytest
 
@@ -256,7 +256,10 @@ def test_list_ips_table(session_test_data: Session, mock_api_client: Mock) -> No
     mock_load_project_or_throw.assert_called_once()
     mock_get_project_id.assert_called_once_with(mock_project_definition.root)
     mock_api_client.list_sessions_api_v2_sessions_get.assert_called_once_with(
-        project_id=session_test_data.project_id, name_match=None, active_only=True
+        project_id=session_test_data.project_id,
+        name_match=None,
+        active_only=True,
+        _request_timeout=ANY,
     )
     mock_get_head_and_worker_ip.assert_called_once_with(
         session_name=session_test_data.name
@@ -315,12 +318,13 @@ def test_list_ips_json(
         project_id=session_test_data.project_id,
         name_match=session_test_data.name,
         active_only=True,
+        _request_timeout=ANY,
     )
     mock_api_client.get_session_commands_history_api_v2_session_commands_get.assert_called_once_with(
         session_id=session_test_data.id
     )
     mock_api_client.get_project_api_v2_projects_project_id_get.assert_called_once_with(
-        project_id=session_test_data.project_id
+        project_id=session_test_data.project_id, _request_timeout=ANY,
     )
     mock_api_client.get_cloud_api_v2_clouds_cloud_id_get.assert_called_once_with(
         cloud_id=session_test_data.cloud_id
