@@ -8,6 +8,8 @@ import pytest
 import yaml
 
 import anyscale
+from anyscale.client.openapi_client.models.application_template import ApplicationTemplate  # type: ignore
+from anyscale.client.openapi_client.models.build import Build  # type: ignore
 from anyscale.client.openapi_client.models.project import Project  # type: ignore
 from anyscale.client.openapi_client.models.project_response import ProjectResponse  # type: ignore
 from anyscale.client.openapi_client.models.session import Session  # type: ignore
@@ -27,6 +29,34 @@ def _make_session(i: int, state: str) -> Session:
         service_proxy_url="http://session-{}.userdata.com/auth?token=value&bar".format(
             i
         ),
+        jupyter_notebook_url="http://session-{}.userdata.com/jupyter/lab?token=value".format(
+            i
+        ),
+    )
+
+
+def _make_app_template() -> ApplicationTemplate:
+    return ApplicationTemplate(
+        project_id="project_id",
+        id="application_template_id",
+        name="test-app-config",
+        creator_id="creator_id",
+        created_at=datetime.now(),
+        last_modified_at=datetime.now(),
+    )
+
+
+def _make_build() -> Build:
+    return Build(
+        id="build_id",
+        revision=0,
+        application_template_id="application_template_id",
+        config_json="",
+        creator_id="creator_id",
+        status="succeeded",
+        created_at=datetime.now(),
+        last_modified_at=datetime.now(),
+        docker_image_name="docker_image_name",
     )
 
 
@@ -50,13 +80,15 @@ def _make_test_builder(
     proj_resp.result.name = "scratch"
     sdk.get_project.return_value = proj_resp
     subprocess = Mock()
+    _os = Mock()
     builder = SessionBuilder(
         scratch_dir=scratch.absolute().as_posix(),
         anyscale_sdk=sdk,
         subprocess=subprocess,
         ray=ray,
+        _os=_os,
     )
-    builder._project_exists = lambda _: False  # type: ignore
+    builder._find_project_id = lambda _: None  # type: ignore
     setattr(builder, "_up_session", Mock())
     setattr(
         builder, "_get_last_used_cloud", Mock(return_value="anyscale_default_cloud")
@@ -124,6 +156,50 @@ def test_fallback_scratch_dir(tmp_path: Path, project_test_data: Project) -> Non
     )
 
 
+def test_background_run_mode(tmp_path: Path, project_test_data: Project) -> None:
+    scratch_dir = (tmp_path / "scratch").absolute().as_posix()
+    builder, sdk, subprocess, ray = _make_test_builder(tmp_path)
+    sdk.create_project.return_value = ProjectResponse(result=project_test_data)
+
+    # Should create a new .anyscale.yaml file in the scratch dir
+    builder.run_mode("background").connect()
+
+    assert anyscale.project.get_project_id(scratch_dir)
+    builder._subprocess.check_output.assert_called_with(
+        ["anyscale", "push", "session-0", "-s", ANY, "-t", ANY], stderr=ANY
+    )
+    builder._subprocess.check_call.assert_called_with(ANY)
+    builder._os._exit.assert_called_once_with(0)
+
+
+def test_local_docker_run_mode(tmp_path: Path, project_test_data: Project) -> None:
+    scratch_dir = (tmp_path / "scratch").absolute().as_posix()
+    builder, sdk, subprocess, ray = _make_test_builder(tmp_path)
+    sdk.create_project.return_value = ProjectResponse(result=project_test_data)
+
+    # Should create a new .anyscale.yaml file in the scratch dir
+    builder.run_mode("local_docker").connect()
+
+    assert anyscale.project.get_project_id(scratch_dir)
+    builder._subprocess.check_call.assert_called_with(
+        [
+            "docker",
+            "run",
+            "--env",
+            ANY,
+            "--env",
+            ANY,
+            "-v",
+            ANY,
+            "--entrypoint=/bin/bash",
+            ANY,
+            "-c",
+            ANY,
+        ]
+    )
+    builder._os._exit.assert_called_once_with(0)
+
+
 def test_connect_with_cloud(tmp_path: Path, project_test_data: Project) -> None:
     scratch_dir = (tmp_path / "scratch").absolute().as_posix()
     builder, sdk, subprocess, ray = _make_test_builder(tmp_path)
@@ -140,7 +216,7 @@ def test_connect_with_cloud(tmp_path: Path, project_test_data: Project) -> None:
 def test_clone_scratch_dir(tmp_path: Path, project_test_data: Project) -> None:
     scratch_dir = (tmp_path / "scratch").absolute().as_posix()
     builder, sdk, subprocess, ray = _make_test_builder(tmp_path)
-    builder._project_exists = lambda _: True
+    builder._find_project_id = lambda _: "foo"
     sdk.create_project.return_value = ProjectResponse(result=project_test_data)
 
     def clone_project(*a: Any, **kw: Any) -> None:
@@ -200,7 +276,7 @@ def test_base_docker_image(tmp_path: Path, project_test_data: Project) -> None:
 
     assert data["docker"]["image"] == "anyscale/ray-ml:custom"
     for nodes_type, node_config in data["available_node_types"].items():
-        assert node_config["docker"]["image"] == "anyscale/ray-ml:custom"
+        assert node_config["docker"]["worker_image"] == "anyscale/ray-ml:custom"
 
 
 def test_requirements_list(tmp_path: Path, project_test_data: Project) -> None:
@@ -296,7 +372,7 @@ def test_reuse_session_hash_match(tmp_path: Path, project_test_data: Project) ->
     os.makedirs(scratch_dir)
     builder.require(["wikipedia", "dask"]).project_dir(scratch_dir)
     cluster_yaml = yaml.safe_load(anyscale.project.CLUSTER_YAML_TEMPLATE)
-    builder._populate_cluster_config(cluster_yaml, "scratch")
+    builder._populate_cluster_config(cluster_yaml, "project_id", "scratch")
     with open(os.path.join(scratch_dir, "session-default.yaml"), "w+") as f:
         f.write(yaml.dump(cluster_yaml))
     local_files_hash = builder._fingerprint(scratch_dir).encode("utf-8")
@@ -458,7 +534,7 @@ def test_fixed_session_not_running(tmp_path: Path, project_test_data: Project) -
     builder.session("session-1").connect()
 
     builder._up_session.assert_called_once_with(
-        ANY, "session-1", scratch_dir,
+        ANY, "session-1", "anyscale_default_cloud", scratch_dir,
     )
 
 
@@ -493,3 +569,78 @@ def test_new_fixed_session(tmp_path: Path, project_test_data: Project) -> None:
     builder._up_session.assert_called_once_with(
         ANY, "session-2", "anyscale_default_cloud", scratch_dir,
     )
+
+
+class MockPopen(object):
+    def __init__(self) -> None:
+        pass
+
+    def communicate(self) -> Tuple[str, str]:
+        return (
+            '[{"id": "cloud2", "name": "second cloud"}, {"id": "cloud1", "name": "first cloud"}]',
+            "",
+        )
+
+
+def test_get_default_cloud(tmp_path: Path, project_test_data: Project) -> None:
+    subprocess = Mock()
+    subprocess.Popen.return_value = MockPopen()
+    sdk = Mock()
+    project_test_data.last_used_cloud_id = None
+    sdk.get_project.return_value = ProjectResponse(result=project_test_data)
+    builder = SessionBuilder(anyscale_sdk=sdk, subprocess=subprocess,)
+    # Check that we get the "default cloud" (cloud first created)
+    # if there is no last used cloud.
+    assert builder._get_last_used_cloud("prj_1") == "first cloud"
+    project_test_data.last_used_cloud_id = "cloud2"
+    # If there is a last used cloud, use that instead.
+    assert builder._get_last_used_cloud("prj_1") == "second cloud"
+
+
+def test_app_config(tmp_path: Path, project_test_data: Project) -> None:
+    scratch_dir = (tmp_path / "scratch").absolute().as_posix()
+    builder, sdk, subprocess, ray = _make_test_builder(tmp_path, session_states=[])
+    sdk.create_project.return_value = ProjectResponse(result=project_test_data)
+
+    app_templates_resp = Mock()
+    app_templates_resp.results = [_make_app_template()]
+    app_templates_resp.metadata.next_paging_token = None
+    sdk.list_application_templates.return_value = app_templates_resp
+
+    build = _make_build()
+    builds_resp = Mock()
+    builds_resp.results = [build]
+    builds_resp.metadata.next_paging_token = None
+    sdk.list_builds.return_value = builds_resp
+
+    get_build_resp = Mock()
+    get_build_resp.result = build
+    sdk.get_build.return_value = get_build_resp
+
+    def create_session(*a: Any, **kw: Any) -> None:
+        sess_resp = Mock()
+        sess_resp.results = [_make_session(0, "Running")]
+        sess_resp.metadata.next_paging_token = None
+        sdk.list_sessions.return_value = sess_resp
+
+    builder._up_session.side_effect = create_session
+
+    with pytest.raises(RuntimeError):
+        builder.app_config("non-existent-app-config").connect()
+
+    builder.app_config("test-app-config").connect()
+
+    with open(
+        (tmp_path / "scratch" / "session-default.yaml").absolute().as_posix()
+    ) as f:
+        data = yaml.safe_load(f)
+
+    assert data["docker"]["image"] == "localhost:5556/docker_image_name"
+    for nodes_type, node_config in data["available_node_types"].items():
+        assert (
+            node_config["docker"]["worker_image"] == "localhost:5556/docker_image_name"
+        )
+
+    builder._up_session.assert_called_once_with(
+        ANY, "session-0", "anyscale_default_cloud", scratch_dir,
+    )
