@@ -1,6 +1,6 @@
 import json
 import os
-from unittest.mock import Mock, mock_open, patch
+from unittest.mock import ANY, Mock, mock_open, patch
 
 from anyscale.client.openapi_client.models.project import Project  # type: ignore
 from anyscale.client.openapi_client.models.project_response import ProjectResponse  # type: ignore
@@ -28,7 +28,10 @@ def test_get_project_sessions(session_test_data: Session) -> None:
 
     assert sessions == [session_test_data]
     mock_api_client.list_sessions_api_v2_sessions_get.assert_called_once_with(
-        project_id=session_test_data.project_id, name_match=None, active_only=True
+        project_id=session_test_data.project_id,
+        name_match=None,
+        active_only=True,
+        _request_timeout=ANY,
     )
 
 
@@ -42,7 +45,10 @@ def test_get_project_session(session_test_data: Session) -> None:
 
     assert session == session_test_data
     mock_api_client.list_sessions_api_v2_sessions_get.assert_called_once_with(
-        project_id=session_test_data.project_id, name_match=None, active_only=True
+        project_id=session_test_data.project_id,
+        name_match=None,
+        active_only=True,
+        _request_timeout=ANY,
     )
 
 
@@ -55,7 +61,7 @@ def test_get_proj_name_from_id(project_test_data: Project) -> None:
 
     assert project_name == project_test_data.name
     mock_api_client.get_project_api_v2_projects_project_id_get.assert_called_once_with(
-        project_id=project_test_data.id
+        project_id=project_test_data.id, _request_timeout=ANY,
     )
 
 
@@ -68,7 +74,7 @@ def test_get_proj_id_from_name(project_test_data: Project) -> None:
 
     assert project_id == project_test_data.id
     mock_api_client.find_project_by_project_name_api_v2_projects_find_by_name_get.assert_called_once_with(
-        name=project_test_data.name
+        name=project_test_data.name, _request_timeout=ANY,
     )
 
 
